#!/usr/bin/env python
# -*- mode: python; coding: koi8-r; -*-
# (c) Con Radchenko mailto:bofh@pochta.ru

import sys, os
from glob import glob
from ConfigParser import ConfigParser

import gtk
from gtk import TRUE, FALSE
from gtk import glade
from gtk import gdk
#import pango

import ornamentbook
from miscutils import entry_dialog, error_dialog, fix_filename, OpenFileDialog
#from detectcharset import detectcharset

glade_file = 'ornamentbook.glade'
config_dir = os.path.expanduser('~/.pybr')
config_file = os.path.expanduser('~/.pybr/ornamentbook.conf')
user_skin_dir = os.path.expanduser('~/.pybr/skins')

##----------------------------------------------------------------------

class NewPyBR:

    def __init__(self, filename=None):

        self.recent_files = []
        # recent_files: list:
        # [
        #   [filename, line_index, char_index, encoding, lang],
        #   ...
        # ]
        self.num_recent_files = 40
        self.filename = None
        self.encoding = 'utf-8'

        self.create_widgets()

        if filename:
            gtk.idle_add(self.open_file, filename)
        elif self.recent_files:
            gtk.idle_add(self.open_file, self.recent_files[-1][0])


    def create_widgets(self):

        gf = None

        for d in sys.path:
            f = os.path.join(d, 'ornamentbook', glade_file)
            if os.path.exists(f):
                gf = f
                break

        if not gf:
            sys.exit('ERROR: can\'t find glade file')

        wTree = glade.XML(gf, 'main_window')
        self.glade_file = gf

        wTree.signal_autoconnect(self)

        self.window = wTree.get_widget('main_window')
        top_vbox = wTree.get_widget('top_vbox')
        text_widget = ornamentbook.OrnamentBook(self.window)
        top_vbox.pack_start(text_widget)

        text_widget.realize()
        #text_widget.connect('position-changed', self.position_changed_cb)
        self.text_widget = text_widget

        self.open_recent_file_menu = wTree.get_widget('open_recent_file_menu')
        self.skin_menu = wTree.get_widget('skin_menu')
        #self.percent_label = wTree.get_widget('percent_label')
        self.hyphenation_menu = wTree.get_widget('hyphenation_menu')

        self.read_properties()

        self.update_skin_menu()
        self.update_hyphenation_menu()

        self.window.show_all()
        #text_widget.grab_focus()


##     def position_changed_cb(self, w, pos):
##         self.percent_label.set_text('%d %%' % int(pos*100))


    def open_file(self, filename):
        #print 'open_file', filename

        try:
            open(filename)
        except IOError, err:
            #sys.stderr.write('ERROR: can\'t open file: %s: %s\n'
            #                 % (filename, err[1]))
            error_dialog(self.window, "Can't open file %s: %s" %
                         (fix_filename(filename), err[1]))
            self.update_recent_files_menu()
            return

        if self.filename: # save previous file
            #print 'save previous file', self.text_widget.current_page_line_index
            self.add_file_to_recent(self.filename,
                                    self.text_widget.current_page_line_index,
                                    self.text_widget.current_page_char_index,
                                    self.text_widget.encoding,
                                    self.text_widget.lang)


        rf = self.get_recent_file(filename)

        if rf:
            ret = self.text_widget.open_file(filename,
                                             rf[3], rf[1], rf[2], rf[4])
        else:
            ret = self.text_widget.open_file(filename)

        if ret: # success
            self.filename = filename
            self.add_file_to_recent(self.filename,
                                    self.text_widget.current_page_line_index,
                                    self.text_widget.current_page_char_index,
                                    self.text_widget.encoding,
                                    self.text_widget.lang)
            self.update_recent_files_menu()



    def read_properties(self):

        try:
            open(config_file)
        except IOError, err:
            sys.stderr.write('ERROR: can\'t open config file: %s: %s\n'
                             % (config_file, err[1]))
            return

        cf = ConfigParser()
        cf.read((config_file,))

        try:
            width = cf.getint('main', 'width')
            height = cf.getint('main', 'height')
            self.window.resize(width, height)
            self.text_widget.left_page_right_margin = cf.getfloat('main', 'left_page_right_margin')
            self.text_widget.right_page_left_margin = cf.getfloat('main', 'right_page_left_margin')
            self.text_widget.left_page_left_margin = cf.getfloat('main', 'left_page_left_margin')
            self.text_widget.right_page_right_margin = cf.getfloat('main', 'right_page_right_margin')
            self.text_widget.top_margin = cf.getfloat('main', 'top_margin')
            self.text_widget.bottom_margin = cf.getfloat('main', 'bottom_margin')
            self.text_widget.set_tab_width(cf.getint('main', 'tabs_width'))
            self.text_widget.set_font(cf.get('main', 'font'))
            self.text_widget.set_fg_color(cf.get('main', 'foreground_color'))
            self.text_widget.set_skin(cf.get('main', 'skin'))

            self.text_widget.single_column = cf.getint('main', 'single_column')

            self.text_widget.percent_x_pos = cf.getfloat('main', 'percent_x_pos')
            self.text_widget.percent_y_pos = cf.getfloat('main', 'percent_y_pos')
            self.text_widget.clock_x_pos = cf.getfloat('main', 'clock_x_pos')
            self.text_widget.clock_y_pos = cf.getfloat('main', 'clock_y_pos')

            OpenFileDialog.file_select_path = cf.get('main', 'file_select_path')
            self.num_recent_files = cf.getint('main', 'num_recent_files')

        except:
            print >> sys.stderr, 'WARNING: error in config file: section "main"'

        try:
            self.text_widget.set_pixels(cf.getfloat('main', 'pixels_inside_wrap'),
                                        cf.getfloat('main', 'pixels_below_line'))
        except:
            print >> sys.stderr, 'WARNING: error in config file: section "main"'

        rf = []
        for i in range(self.num_recent_files):
            section = 'recent%d' % i
            if cf.has_section(section):
                try:
                    rf.append((cf.get(section, 'filename'),
                               cf.getint(section, 'line'),
                               cf.getint(section, 'index'),
                               cf.get(section, 'encoding'),
                               cf.get(section, 'lang')))
                except:
                    print >> sys.stderr, \
                          'WARNING: error in config file: section "%s"' % section
        self.recent_files = rf



    def save_properties(self):

        try:
            fd = open(config_file, 'w')
        except IOError, err:
            sys.stderr.write('ERROR: can\'t open config file: %s: %s'
                             % (config_file, err[1]))
            return

        cf = ConfigParser()
        cf.add_section('main')

        width, height = self.window.get_size()
        cf.set('main', 'width', width)
        cf.set('main', 'height', height)
        if self.filename:
            self.add_file_to_recent(self.filename,
                                    self.text_widget.current_page_line_index,
                                    self.text_widget.current_page_char_index,
                                    self.text_widget.encoding,
                                    self.text_widget.lang)


        cf.set('main', 'left_page_left_margin', self.text_widget.left_page_left_margin)
        cf.set('main', 'left_page_right_margin', self.text_widget.left_page_right_margin)
        cf.set('main', 'right_page_left_margin', self.text_widget.right_page_left_margin)
        cf.set('main', 'right_page_right_margin', self.text_widget.right_page_right_margin)
        cf.set('main', 'top_margin', self.text_widget.top_margin)
        cf.set('main', 'bottom_margin', self.text_widget.bottom_margin)
        cf.set('main', 'tabs_width', self.text_widget.tabs_width)
        cf.set('main', 'single_column', self.text_widget.single_column)
        cf.set('main', 'percent_x_pos', self.text_widget.percent_x_pos)
        cf.set('main', 'percent_y_pos', self.text_widget.percent_y_pos)
        cf.set('main', 'clock_x_pos', self.text_widget.clock_x_pos)
        cf.set('main', 'clock_y_pos', self.text_widget.clock_y_pos)
        pixels = self.text_widget.get_pixels()
        cf.set('main', 'pixels_inside_wrap', pixels[0])
        cf.set('main', 'pixels_below_line', pixels[1])

        color = '#%02x%02x%02x' % (self.text_widget._color.red>>8,
                                   self.text_widget._color.green>>8,
                                   self.text_widget._color.blue>>8)
        cf.set('main', 'foreground_color', color)

        font = self.text_widget.get_font()
        cf.set('main', 'font', font)

        cf.set('main', 'skin', self.text_widget.skin_file)
        cf.set('main', 'file_select_path', OpenFileDialog.file_select_path)
        cf.set('main', 'num_recent_files', self.num_recent_files)

        i = 0
        for filename, line, index, encoding, lang in self.recent_files:
            section = 'recent%d' % i
            cf.add_section(section)
            cf.set(section, 'filename', filename)
            cf.set(section, 'line', line)
            cf.set(section, 'index', index)
            cf.set(section, 'encoding', encoding)
            cf.set(section, 'lang', lang)
            i+=1

        cf.write(fd)


    def update_recent_files_menu(self):
        menu = gtk.Menu()
        menu.show()
        for f in self.recent_files:
            menuitem = gtk.MenuItem(fix_filename(f[0]))
            menuitem.get_child().set_use_underline(FALSE)
            menuitem.connect('activate', self.recent_files_activate_cb, f[0])
            menu.add(menuitem)
            menuitem.show()
        self.open_recent_file_menu.set_submenu(menu)


    def update_hyphenation_menu(self):
        menu = gtk.Menu()
        menu.show()
        from hyphenation import Hyphenation
        langs = Hyphenation().get_langs()
        for l, n in langs:
            menuitem = gtk.MenuItem(n)
            menuitem.get_child().set_use_underline(FALSE)
            menuitem.connect('activate', self.hyphenation_menu_activate_cb, l)
            menu.add(menuitem)
            menuitem.show()
        self.hyphenation_menu.set_submenu(menu)


    def update_skin_menu(self):

        skins = []
        if os.path.exists(user_skin_dir) and os.path.isdir(user_skin_dir):
            skins_dirs = [user_skin_dir]
        else:
            skins_dirs = []

        for d in sys.path:
            f = os.path.join(d, 'ornamentbook', ornamentbook.skins_dir)
            if os.path.exists(f) and os.path.isdir(f):
                skins_dirs.append(f)
                break

        for d in skins_dirs:
            f = os.path.join(d, '*.jpg')
            skins.extend(glob(f))

        menu = gtk.Menu()
        menu.show()
        for f in skins:
            fn = os.path.abspath(f)
            menuitem = gtk.MenuItem(fn)
            menuitem.get_child().set_use_underline(FALSE)
            menuitem.connect('activate', self.skin_menu_activate_cb, fn)
            menu.add(menuitem)
            menuitem.show()
        self.skin_menu.set_submenu(menu)


    def skin_menu_activate_cb(self, w, filename):
        #print 'skin_menu_activate_cb', filename
        self.text_widget.set_skin(filename)
        self.text_widget.draw_text()


    def hyphenation_menu_activate_cb(self, w, lang):
        self.text_widget.lang = lang
        self.text_widget.draw_text()


    def recent_files_activate_cb(self, w, filename):
        #print 'recent_files_activate_cb', filename
        self.open_file(filename)


    def get_recent_file(self, filename):
        #print 'get_recent_file', filename
        for f in self.recent_files:
            if f[0] == filename:
                return f
        return None


    def add_file_to_recent(self, filename, line, index, encoding, lang):

        i = 0
        for f in self.recent_files:
            if f[0] == filename:
                del self.recent_files[i]
                break
            i+=1
        if len(self.recent_files) >= self.num_recent_files:
            del self.recent_files[:len(self.recent_files)-self.num_recent_files+1]
        self.recent_files.append((filename, line, index, encoding, lang))

        #self.update_recent_files_menu()


    def on_open_activate(self, *args):

        dialog = OpenFileDialog(self.window)
        filename = dialog.filename
        if filename:
            self.open_file(filename)
        return


    def on_go_to_position_activate(self, *args):
        s = entry_dialog(self.window, 'Go to position', 'Go to position: (%) ')
        try:
            pos = float(s)
        except:
            return
        if 100 < pos < 0:
            return
        self.text_widget.set_position(pos/100)


    def on_quit_activate(self, *args):
        self.window.destroy()


    def on_main_window_unrealize(self, *args):
        #print 'on_main_window_unrealize'
        self.save_properties()

    def on_properties_activate(self, *args):
        #print 'on_properties_activate'
        wTree = glade.XML(self.glade_file, 'properties_dialog')
        # move dialog to centre of parrent window
        dialog = wTree.get_widget('properties_dialog')
        dialog.set_transient_for(self.window)
        x, y = self.window.get_position()
        w, h = self.window.get_size()
        dw, dh = dialog.get_size()
        dx = x + (w-dw)/2
        dy = y + (h-dh)/2
        dialog.move(dx, dy)
        dialog.show()
        dic = {
            'on_close_properties_clicked' :
            (lambda w: dialog.destroy()),
            'on_font_select_clicked' :
            (lambda w: self.on_font_select_clicked(dialog, 'font')),
            'on_foreground_color_clicked' :
            (lambda w: self.on_color_select_clicked(dialog, 'foreground_color')),
            'on_apply_properties_clicked' :
            self.apply_properties_cb,
            }
        wTree.signal_autoconnect(dic)

        self.single_column_checkbutton \
            = wTree.get_widget('single_column_checkbutton')

        self.left_page_left_margin_spinbutton \
             = wTree.get_widget('left_page_left_margin_spinbutton')
        self.left_page_right_margin_spinbutton \
             = wTree.get_widget('left_page_right_margin_spinbutton')
        self.right_page_left_margin_spinbutton \
             = wTree.get_widget('right_page_left_margin_spinbutton')
        self.right_page_right_margin_spinbutton \
             = wTree.get_widget('right_page_right_margin_spinbutton')

        self.top_margin_spinbutton \
             = wTree.get_widget('top_margin_spinbutton')
        self.bottom_margin_spinbutton \
             = wTree.get_widget('bottom_margin_spinbutton')

        self.tabs_width_spinbutton = wTree.get_widget('tabs_width_spinbutton')

        self.percent_x_pos_spinbutton \
            = wTree.get_widget('percent_x_pos_spinbutton')
        self.percent_y_pos_spinbutton \
            = wTree.get_widget('percent_y_pos_spinbutton')
        self.clock_x_pos_spinbutton = wTree.get_widget('clock_x_pos_spinbutton')
        self.clock_y_pos_spinbutton = wTree.get_widget('clock_y_pos_spinbutton')

        self.pixels_inside_wrap_spinbutton \
            = wTree.get_widget('pixels_inside_wrap_spinbutton')
        self.pixels_below_line_spinbutton \
            = wTree.get_widget('pixels_below_line_spinbutton')

        self.single_column_checkbutton.set_active(self.text_widget.single_column)

        self.left_page_left_margin_spinbutton.set_value(self.text_widget.left_page_left_margin)
        self.left_page_right_margin_spinbutton.set_value(self.text_widget.left_page_right_margin)
        self.right_page_left_margin_spinbutton.set_value(self.text_widget.right_page_left_margin)
        self.right_page_right_margin_spinbutton.set_value(self.text_widget.right_page_right_margin)

        self.top_margin_spinbutton.set_value(self.text_widget.top_margin)
        self.bottom_margin_spinbutton.set_value(self.text_widget.bottom_margin)

        self.tabs_width_spinbutton.set_value(self.text_widget.tabs_width)

        self.percent_x_pos_spinbutton.set_value(self.text_widget.percent_x_pos)
        self.percent_y_pos_spinbutton.set_value(self.text_widget.percent_y_pos)
        self.clock_x_pos_spinbutton.set_value(self.text_widget.clock_x_pos)
        self.clock_y_pos_spinbutton.set_value(self.text_widget.clock_y_pos)

        pixels = self.text_widget.get_pixels()
        self.pixels_inside_wrap_spinbutton.set_value(pixels[0])
        self.pixels_below_line_spinbutton.set_value(pixels[1])




    def apply_properties_cb(self, *args):
        #print 'apply_properties_cb'

        self.text_widget.left_page_left_margin = self.left_page_left_margin_spinbutton.get_value()
        self.text_widget.left_page_right_margin = self.left_page_right_margin_spinbutton.get_value()
        self.text_widget.right_page_left_margin = self.right_page_left_margin_spinbutton.get_value()
        self.text_widget.right_page_right_margin = self.right_page_right_margin_spinbutton.get_value()
        self.text_widget.top_margin = self.top_margin_spinbutton.get_value()
        self.text_widget.bottom_margin = self.bottom_margin_spinbutton.get_value()
        self.text_widget.set_tab_width(int(self.tabs_width_spinbutton.get_value()))
        self.text_widget.single_column = self.single_column_checkbutton.get_active()


        self.text_widget.percent_x_pos = self.percent_x_pos_spinbutton.get_value()
        self.text_widget.percent_y_pos = self.percent_y_pos_spinbutton.get_value()
        self.text_widget.clock_x_pos = self.clock_x_pos_spinbutton.get_value()
        self.text_widget.clock_y_pos = self.clock_y_pos_spinbutton.get_value()
        self.text_widget.set_pixels(self.pixels_inside_wrap_spinbutton.get_value(),
                                    self.pixels_below_line_spinbutton.get_value())

        self.text_widget.draw_text()



    def on_font_select_clicked(self, parent, prop):
        #print 'on_font_select_clicked'
        dialog = gtk.FontSelectionDialog('Select font')
        dialog.set_font_name(self.text_widget.get_font())
        dialog.set_transient_for(parent)
        dialog.show()

        response = dialog.run()
        if response  == gtk.RESPONSE_OK:
            font = dialog.get_font_name()
            self.text_widget.set_font(font)
            self.text_widget.set_tab_width()
            self.text_widget.draw_text()
            #props = {prop: font}
            #self.apply_properties(props)

        dialog.destroy()


    def on_color_select_clicked(self, parent, prop):
        #print 'on_color_clicked'
        dialog = gtk.ColorSelectionDialog('Select color')

        dialog.colorsel.set_has_palette(TRUE)
        dialog.set_transient_for(parent)
        dialog.show()

        response = dialog.run()

        if response == gtk.RESPONSE_OK:

            color = dialog.colorsel.get_current_color()
            self.text_widget.set_fg_color(color)
            self.text_widget.draw_text()
            #color_val = '#%02x%02x%02x' \
            #            % (color.red>>8, color.green>>8, color.blue>>8)
            #props = {prop: color_val}
            #self.apply_properties(props)

        dialog.destroy()


    def on_increase_font_activate(self, *args):
        self.text_widget.incr_font(1)
        self.text_widget.draw_text()


    def on_decrease_font_activate(self, *args):
        self.text_widget.incr_font(-1)
        self.text_widget.draw_text()


##----------------------------------------------------------------------

if __name__ == '__main__':
    try:
        os.mkdir(config_dir)
    except:
        pass


    npbr = NewPyBR()
    npbr.window.connect('destroy', lambda w: gtk.main_quit())
    gtk.main()


