#!/usr/bin/env python
# -*- mode: python; coding: koi8-r -*-

# from __future__ imports must occur at the beginning of the file
from __future__ import generators

import string, sys, os
import tempfile
import time
import base64
import types

import gobject, gtk, pango
#from gtk import TRUE, FALSE
from gtk import gdk

from detectcharset import detectcharset
from textparagraph import TextParagraph
from hyphenation import Hyphenation
from miscutils import error_dialog, fix_filename

skins_dir = 'skins'

PANGO_SCALE = pango.SCALE
#PANGO_SCALE = 1024

default_lang = 'ru'



class OrnamentBook(gtk.DrawingArea):

    keysyms_next = (gtk.keysyms.KP_Page_Down,
                    gtk.keysyms.Page_Down,
                    gtk.keysyms.space,
                    gtk.keysyms.Right)
    keysyms_prev = (gtk.keysyms.KP_Page_Up,
                    gtk.keysyms.Page_Up,
                    gtk.keysyms.Left)

    def __init__(self, parent):

        default_skin = 'golden.jpg'
        skin_file = None
        for d in sys.path:
            f = os.path.join(d, 'ornamentbook', skins_dir, default_skin)
            if os.path.exists(f):
                skin_file = f
                break
        if not skin_file:
            sys.exit('ERROR: can\'t find default skin file %s' % default_skin)
        self.skin_file = skin_file

        gtk.DrawingArea.__init__(self)

        # Hyphenation
        self.hyph = Hyphenation()
        self.lang = default_lang

        self._parent = parent
        self.gc = None
        #self.area = gtk.DrawingArea()

##         self.set_size_request(950, 400)
##         self._width = 950
##         self._height = 400

        gobject.signal_new('position-changed', OrnamentBook,
                           gobject.SIGNAL_RUN_LAST,
                           gobject.TYPE_NONE,
                           (gobject.TYPE_FLOAT,))

        self.set_size_request(600, 400)
        self._width = 600
        self._height = 400

        #self.set_events(gdk.ALL_EVENTS_MASK)
        #self.add_events(gdk.BUTTON_PRESS_MASK|gdk.BUTTON_RELEASE_MASK)
        #self.add_events(gdk.BUTTON_RELEASE_MASK)
        self.add_events(gdk.BUTTON_PRESS_MASK)
        #self.add_events(gdk.BUTTON_PRESS_MASK|gdk.BUTTON_RELEASE_MASK|gdk.KEY_PRESS_MASK|gdk.KEY_RELEASE_MASK)
        #self.set_events(gdk.ALL_EVENTS_MASK)

        self.connect('expose-event', self.area_expose_cb)
        #self.connect('button_release_event', self.button_release_event_cb)
        self.connect('button_press_event', self.button_release_event_cb)
        #self.realize()

        #self._parent.connect('key_press_event', self.key_press_event_cb)
        self._parent.connect('key-release-event', self.key_release_event_cb)

        self.current_page_line_index = 0
        self.current_page_char_index = 0

        self.next_page_line_index = 0
        self.next_page_char_index = 0

        self.pixmap_area = None
        #self.prev_pixmap_area = None
        self.next_pixmap_area = None

        self.pixels_inside_wrap = 0
        self.pixels_below_line = 0

        self.set_font('Sans 14')
        self._color = gdk.color_parse('#000000')

        self.top_margin = 10
        self.bottom_margin = 60
        self.left_page_left_margin = 30
        self.left_page_right_margin = 30
        self.right_page_left_margin = 30
        self.right_page_right_margin = 30
        self.set_tab_width(5)
        self.percent_x_pos = 30
        self.percent_y_pos = 20
        self.clock_x_pos = 30
        self.clock_y_pos = 20
        self.single_column = 0
        self.show_status = 0
        self.footnote_line_width = 2

        self.content = []
        self.text_length = 0

        self.footnotes_list = []
        self.prev_index_list = []

        return


    def set_skin(self, filename):
        try:
            open(filename)
        except IOError, err:
            sys.stderr.write('ERROR: can\'t open file: %s: %s\n'
                             % (filename, err[1]))
            return
        self.skin_file = filename
        self.pixbuf = gdk.pixbuf_new_from_file(self.skin_file)
        self.cur_pixbuf = self.pixbuf.scale_simple(self._width, self._height,
                                                   gdk.INTERP_BILINEAR)


    def set_font(self, font):

        self.font = pango.FontDescription(font)
        pangolayout = self.create_pango_layout(' ')
        pangolayout.set_font_description(self.font)
        self.space_width = pangolayout.get_extents()[1][2]/PANGO_SCALE
        self.text_height = pangolayout.get_extents()[1][3]/PANGO_SCALE \
                           + self.pixels_inside_wrap
        self.text_par_height = pangolayout.get_extents()[1][3]/PANGO_SCALE \
                               + self.pixels_below_line

        self.title_font = self.font.copy()
        self.title_font.set_weight(pango.WEIGHT_BOLD)

        self.footnote_font = self.font.copy()
        self.footnote_font.set_size(self.font.get_size()-2*PANGO_SCALE)
        pangolayout = self.create_pango_layout(' ')
        pangolayout.set_font_description(self.footnote_font)
        self.footnote_space_width = pangolayout.get_extents()[1][2]/PANGO_SCALE
        self.footnote_height = pangolayout.get_extents()[1][3]/PANGO_SCALE \
                               + self.pixels_inside_wrap
        self.footnote_par_height = pangolayout.get_extents()[1][3]/PANGO_SCALE \
                                   + self.pixels_below_line

        if self.window:
            x, y, widget_width, widget_height, depth = self.window.get_geometry()
            self.status_pixmap_area = gdk.Pixmap(None,
                                                 self._width,
                                                 self.text_height+self.percent_y_pos,
                                                 depth)


    def get_font(self):
        return self.font.to_string()


    def incr_font(self, incr):
        # if incr > 0 -> increase, else -> decrease

        fsize = self.font.get_size() # in PANGO_SCALE units

        if incr > 0:
            if fsize < 14*PANGO_SCALE:
                fsize += 1*PANGO_SCALE
            elif fsize < 28*PANGO_SCALE:
                fsize += 2*PANGO_SCALE
            elif fsize < 40*PANGO_SCALE:
                fsize += 4*PANGO_SCALE
            else:
                fsize += 8*PANGO_SCALE
        else:
            if fsize > 40*PANGO_SCALE:
                fsize -= 8*PANGO_SCALE
            elif fsize > 28*PANGO_SCALE:
                fsize -= 4*PANGO_SCALE
            elif fsize > 14*PANGO_SCALE:
                fsize -= 2*PANGO_SCALE
            else:
                fsize -= 1*PANGO_SCALE

        if 8*PANGO_SCALE <= fsize <= 72*PANGO_SCALE:
            self.font.set_size(fsize)
            self.set_font(self.font.to_string())


    def set_pixels(self, pixels_inside_wrap=None, pixels_below_line=None):
        if pixels_inside_wrap is not None:
            self.text_height += pixels_inside_wrap-self.pixels_inside_wrap
            self.footnote_height += pixels_inside_wrap-self.pixels_inside_wrap
            self.pixels_inside_wrap = pixels_inside_wrap

        if pixels_below_line is not None:
            self.text_par_height += pixels_below_line-self.pixels_below_line
            self.footnote_par_height += pixels_below_line-self.pixels_below_line
            self.pixels_below_line = pixels_below_line


    def get_pixels(self):
        return self.pixels_inside_wrap, self.pixels_below_line


    def set_fg_color(self, color):
        if type(color) is types.StringType:
            self._color = gdk.color_parse(color)
        else:
            self._color = color
        self.get_colormap().alloc_color(self._color)
        if self.gc:
            self.gc.set_foreground(self._color)


    def set_tab_width(self, tab_width=None):
        if tab_width is None:
            tab_width = self.tabs_width
        pl = self.create_pango_layout(' '*tab_width)
        width, height = pl.get_pixel_size()

        ta = pango.TabArray(1, True)
        ta.set_tab(0, pango.TAB_LEFT, width)
        self.tab_array = ta
        self.tabs_width = tab_width


    def open_file(self, filename, encoding=None, line=0, index=0, lang=None):

        self.encoding = encoding

        if filename[-4:] == '.fb2':

            ret = self.open_fb2_file(filename, line, index, lang)

        else:

            if not encoding:
                fd = open(filename)
                enc = detectcharset(fd)

                self.encoding = enc or 'iso8859-1' # FixMe

            ret = self.open_plain_text_file(filename, self.encoding,
                                            line, index, lang)

        if ret:
            self.text_length = reduce(lambda x, y: x+len(y.data),
                                      self.content, 0)
            self.draw_text()

        return ret


    def open_plain_text_file(self, filename, encoding,
                             line_index=0, char_index=0,
                             lang=None):

        try:
            text = open(filename).read()
        except IOError, err:
            error_dialog(self._parent, "Can't open file %s: %s" %
                         (fix_filename(filename), err[1]))
            return False

        # remove \x00 chars
        text = text.replace('\0', '')

        text = unicode(text, encoding)

        self.content = []
        for l in text.splitlines():
            paragraph = TextParagraph('text')
            paragraph.data = l
            self.content.append(paragraph)

        self.lang = lang or default_lang

        self.current_page_line_index = line_index
        self.current_page_char_index = char_index

        return True


    def open_fb2_file(self, filename, line_index=0, char_index=0, lang=None):

        from fb2parser import FB2Parser
        from libxml2 import parserError

        self.encoding = 'utf-8'

        fb = FB2Parser()
        try:
            fb.parseFB2(filename)
        except parserError, err:
            #print err
            error_dialog(self._parent, 'Can\'t parse XML file %s: %s'
                         % (fix_filename(filename), str(err)))
            return False

        self.fb = fb
        self.content = fb.content

        if lang:
            self.lang = lang
        else:
            self.lang = fb.lang

        self.current_page_line_index = line_index
        self.current_page_char_index = char_index

        return True


    def realize(self):
        gtk.DrawingArea.realize(self)
        self.gc = self.window.new_gc(line_width=self.footnote_line_width,
                                     foreground=self._color)
        self.pixbuf = gdk.pixbuf_new_from_file(self.skin_file)
        self.cur_pixbuf = self.pixbuf.scale_simple(self._width, self._height,
                                                   gdk.INTERP_BILINEAR)

##         width = self._width
##         height = self.text_height+self.percent_y_pos
##         pb = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8,
##                         self._width,
##                         self.text_height+self.percent_y_pos)

##         self.status_pixmap_area = gdk.Pixmap(None,
##                                              widget_width,
##                                              widget_height,
##                                              depth)

        gtk.timeout_add(10000, self.draw_status, self.window)


    def area_expose_cb(self, area, event):
        x, y, w, h, d = self.window.get_geometry()
        if self.pixmap_area and self._width == w and self._height == h:
            self.window.draw_drawable(self.gc, self.pixmap_area,
                                      0, 0, 0, 0, w, h)
            #self.draw_status() #   ;)
        else:
            self.draw_text()


    def draw_text(self, go_fore=False):
        #print 'draw_text', go_fore

        if not go_fore:
            #  
            self.footnotes_list = []
            self.prev_index_list = []


        x, y, widget_width, widget_height, depth = self.window.get_geometry()

        if not self.pixmap_area \
               or self._width != widget_width \
               or self._height != widget_height:
            self.cur_pixbuf = self.pixbuf.scale_simple(widget_width,
                                                       widget_height,
                                                       gdk.INTERP_BILINEAR)
            self._width = widget_width
            self._height = widget_height

            self.pixmap_area = gdk.Pixmap(None,
                                          widget_width,
                                          widget_height,
                                          depth)

##             self.prev_pixmap_area = gdk.Pixmap(None,
##                                                widget_width,
##                                                widget_height,
##                                                depth)
            self.next_pixmap_area = gdk.Pixmap(None,
                                               widget_width,
                                               widget_height,
                                               depth)

            #status_w = self._width
            #status_h = self.text_height+self.percent_y_pos
            self.status_pixmap_area = gdk.Pixmap(None,
                                          self._width,
                                          self.text_height+self.percent_y_pos,
                                          depth)

##             self.cur_pixbuf.render_to_drawable(self.status_pixmap_area,
##                                                self.gc, 0, 0,
##                                                0, 0,
##                                                status_w,
##                                                status_h,
##                                                False, 0, 0)

        # current page
        if not go_fore: #self.pixmap_area:
            line_index, index = self._draw(self.pixmap_area,
                                           self.current_page_line_index,
                                           self.current_page_char_index)
            self.next_page_line_index = line_index
            self.next_page_char_index = index
            #self.draw_status(self.pixmap_area)
            #self.window.draw_drawable(self.gc, self.pixmap_area,
            #                          0, 0, 0, 0, widget_width, widget_height)

        else:
            self.next_page_line_index = self.next_next_page_line_index
            self.next_page_char_index = self.next_next_page_char_index
            #self.pixmap_area = self.next_pixmap_area
            self.pixmap_area, self.next_pixmap_area = \
                              self.next_pixmap_area, self.pixmap_area


        self.draw_status(self.pixmap_area)

        self.window.draw_drawable(self.gc, self.pixmap_area,
                                  0, 0, 0, 0, widget_width, widget_height)

        # next page
        line_index, index = self._draw(self.next_pixmap_area,
                                       self.next_page_line_index,
                                       self.next_page_char_index)
        self.next_next_page_line_index = line_index
        self.next_next_page_char_index = index


    def draw_status(self, da):

        width = self._width
        height = self.text_height+self.percent_y_pos

##         fn = os.path.expanduser('~/.pybr/skins/my-status.jpg')

##         pixbuf = gdk.pixbuf_new_from_file(fn)
##         cur_pixbuf = self.pixbuf.scale_simple(self._width, self._height,
##                                               gdk.INTERP_BILINEAR)

        #gdk.Pixbuf(colorspace, has_alpha, bits_per_sample, width, height)

##         pb = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, width, height)
##         self.cur_pixbuf.copy_area(0,
##                                   self._height-height,
##                                   width,
##                                   height,
##                                   pb, 0, 0)
##         pixmap, mask = pb.render_pixmap_and_mask()


        self.cur_pixbuf.render_to_drawable(self.status_pixmap_area,
                                           self.gc,
                                           0, self._height-height,
                                           0, 0,
                                           width,
                                           height,
                                           False, 0, 0)

        # position
        pos = self.calc_position()
        pangolayout = self.create_pango_layout(str(int(pos*100))+'%')
        pangolayout.set_font_description(self.font)
        self.status_pixmap_area.draw_layout(self.gc,
                           self.percent_x_pos, 0,
                           pangolayout,
                           foreground=self._color)

        # time
        t = time.strftime('%H:%M', time.localtime())
        pangolayout = self.create_pango_layout(t)
        pangolayout.set_font_description(self.font)
        x_pos = width-self.clock_x_pos \
                -pangolayout.get_extents()[1][2]/PANGO_SCALE
        self.status_pixmap_area.draw_layout(self.gc,
                           x_pos,
                           0,
                           pangolayout,
                           foreground=self._color)
                           #width-self.clock_x_pos,


        da.draw_drawable(self.gc, self.status_pixmap_area,
                         0, 0, 0, self._height-height,
                         width, height)

##         self.window.draw_drawable(self.gc, pixmap,
##                                   0, 0,
##                                   0, self._height-height,
##                                   width, height)

##         self.pixmap_area.draw_drawable(self.gc, pixmap,
##                                   0, 0,
##                                   0, self._height-height,
##                                   width, height)



        return True


    def _draw(self, area_window, line_index, index):

        def draw_1(line_index, index, x_offset, y_offset, text_area_width):

            if index: # draw tail of line if need

                y_offset, index = self.draw_paragraph(area_window,
                                                      line_index, index,
                                                      x_offset, y_offset,
                                                      text_area_width,
                                                      text_area_height)

                if index == -1:
                    index = 0
                    line_index += 1

            elif len(self.content) > 0:

                #      
                while True:
                    if len(self.content) <= line_index or \
                           self.content[line_index].type != 'text' or \
                           self.content[line_index].data != '':
                        break
                    line_index += 1


            if y_offset+self.text_par_height <= self.top_margin+text_area_height:

                for par in self.content[line_index:]:

                    if par.type == 'image':

                        yy = self.draw_image(area_window,
                                             self.fb.images[par.href],
                                             x_offset, y_offset,
                                             text_area_width, text_area_height)
                        if yy >= 0:
                            index = -1
                            y_offset = yy
                        else:
                            break

                    else:
                        y_offset, index = self.draw_paragraph(area_window,
                                                              line_index,
                                                              0,
                                                              x_offset,
                                                              y_offset,
                                                              text_area_width,
                                                              text_area_height)

                    if index == -1:
                        index = 0
                        line_index += 1

                    if y_offset+self.text_par_height > self.top_margin+text_area_height:
                        break

            return line_index, index


        widget_width = self._width
        widget_height = self._height

        self.cur_pixbuf.render_to_drawable(area_window, self.gc, 0, 0,
                                           0, 0, -1, -1, False, 0, 0)

        text_area_height = widget_height-self.top_margin-self.bottom_margin

        # debug
        #area_window.draw_line(self.gc, 0, widget_height-self.bottom_margin,
        #                      widget_width, widget_height-self.bottom_margin)

        if self.single_column:

            x_offset = self.left_page_left_margin
            y_offset = self.top_margin
            text_area_width = widget_width-self.left_page_left_margin \
                              -self.left_page_right_margin
            line_index, index = draw_1(line_index, index,
                                       x_offset, y_offset, text_area_width)

        else:

            # draw left page
            x_offset = self.left_page_left_margin
            y_offset = self.top_margin
            text_area_width = widget_width/2-self.left_page_left_margin \
                              -self.left_page_right_margin
            line_index, index = draw_1(line_index, index,
                                       x_offset, y_offset, text_area_width)

            # draw right page
            x_offset = widget_width/2+self.right_page_left_margin
            y_offset = self.top_margin
            text_area_width = widget_width/2-self.right_page_left_margin \
                              -self.right_page_right_margin
            line_index, index = draw_1(line_index, index,
                                       x_offset, y_offset, text_area_width)

        #self.next_page_line_index = line_index
        #self.next_page_char_index = index

        #area_window.thaw_updates()

        return line_index, index



    def get_pixbuf(self, data, max_width): #, max_height):

        data = base64.decodestring(data)
        tmp_file = tempfile.mktemp()
        try:
            open(tmp_file, 'w').write(data)
            pb = gdk.pixbuf_new_from_file(tmp_file)
            #pb = gtk.gdk.pixbuf_new_from_inline(len(data), data, TRUE)
        finally:
            os.remove(tmp_file)
        w = pb.get_width()
        h = pb.get_height()
        if w > max_width:

            pb = pb.scale_simple(max_width,
                                 h*float(max_width)/w,
                                 gdk.INTERP_BILINEAR)
        return pb


    def draw_image(self, area_window,
                   data,
                   x_offset,
                   y_offset,
                   text_area_width,
                   text_area_height):

##         data = base64.decodestring(data)
##         tmp_file = tempfile.mktemp()
##         try:
##             open(tmp_file, 'w').write(data)
##             pb = gdk.pixbuf_new_from_file(tmp_file)
##             #pb = gtk.gdk.pixbuf_new_from_inline(len(data), data, TRUE)
##             #self.binary_image.set_from_pixbuf(pb)
##         finally:
##             os.remove(tmp_file)

        pb = self.get_pixbuf(data, text_area_width)
        is_top = y_offset==self.top_margin and 1 or 0

        width = pb.get_width()
        height = pb.get_height()

        if not is_top and y_offset+height > text_area_height:
            return -1
        else:
            #pb = pb.add_alpha(TRUE, chr(255), chr(255), chr(255)) # 
            pb.render_to_drawable(area_window, self.gc, 0, 0,
                                  x_offset+(text_area_width-width)/2, y_offset,
                                  width, height,
                                  False, 0, 0)

##             render_to_drawable_alpha(drawable,
##                                      src_x, src_y,
##                                      dest_x, dest_y,
##                                      width, height,
##                                      alpha_mode, alpha_threshold,
##                                      dither, x_dither, y_dither)

##             pb.render_to_drawable_alpha(area_window, 0, 0,
##                                         x_offset+(text_area_width-width)/2,
##                                         y_offset,
##                                         width, height,
##                                         0, 0,
##                                         FALSE, 0, 0)

            return y_offset+height



    def draw_paragraph(self, area_window,
                       line_index, char_index,
                       x_offset, y_offset,
                       text_area_width, text_area_height):

        #      -  -1

        new_index = 0
        y = y_offset

        while True:

            footnotes_height = 0.0
            for ft in self.footnotes_list:
                if ft == []:
                    footnotes_height += self.footnote_par_height-self.footnote_height
                else:
                    footnotes_height += self.footnote_height

            if footnotes_height and \
                   y+footnotes_height+self.text_height >= text_area_height:
                self.draw_footnotes(area_window, x_offset, y,
                                    text_area_width, text_area_height)
                y = self._height
                new_index = char_index
                break

            if y+self.text_height > text_area_height+self.top_margin:
                break

            if not self.content[line_index].data:
                # empty line
                y += self.text_par_height
                new_index = -1
                break

            is_title = self.content[line_index].type in ('title', 'subtitle') \
                       and True or False

            font = is_title and self.title_font or self.font

            pl, new_index = self.get_pangolayout_list(text_area_width, font,
                                                      self.content[line_index],
                                                      char_index)

            text_width = reduce(lambda a, b: a+b, map(lambda x: x[1], pl), 0)

            if is_title:
                space = self.space_width
            elif len(pl) <= 1:
                # single word
                space = 0
            elif new_index == 0:
                # last line in this paragraph
                space = self.space_width
            else:
                space = (text_area_width-text_width)/(len(pl)-1)

            if is_title:
                p_length = 0
                for p in pl:
                    p_length += p[1]
                x = x_offset + (text_area_width - (len(pl)-1)*space - p_length)/2
            else:
                x = x_offset


            for i in range(len(pl)):

                if space and i == len(pl)-1 and not new_index == 0 and not is_title:
                    # last word in this line
                    x = x_offset+text_area_width-pl[i][1]

                # -  
                yy = y+self.text_height-pl[i][0].get_extents()[1][3]/PANGO_SCALE
                area_window.draw_layout(self.gc, x, yy, pl[i][0],
                                        foreground=self._color)

##                 area_window.draw_layout(self.gc, x, y, pl[i][0],
##                                         foreground=self._color)



                x += pl[i][1]+space

            if new_index == 0:
                # new paragraph
                y += self.text_par_height
            else:
                y += self.text_height

            # -- footnotes --
            if self.content[line_index].footnotes:
                ni = new_index == 0 \
                     and len(self.content[line_index].data) \
                     or  new_index
                footnotes_list = self.create_footnotes_list(text_area_width,
                                                            line_index,
                                                            char_index, ni)
                if footnotes_list:
                    self.footnotes_list.extend(footnotes_list)

            if new_index == 0:
                new_index = -1
                break
            else:
                char_index = new_index

        return y, new_index


    def draw_footnotes(self, area_window,
                       x_offset, y, text_area_width, text_area_height):

        if y+self.text_par_height > text_area_height+self.top_margin:
            print '++', y+self.text_par_height, text_area_height+self.top_margin
            return

        list_index = 0

        if gtk.pygtk_version[0] == 2 and gtk.pygtk_version[1] == 0:
            # workarownd for pygtk version < 2.2
            gc = area_window.new_gc(line_width=self.footnote_line_width)
        else:
            gc = self.gc

        area_window.draw_line(gc, x_offset, y+self.footnote_height*2/3,
                              x_offset+text_area_width/3,
                              y+self.footnote_height*2/3)
        y += self.footnote_height

        for pl in self.footnotes_list:

            if not self.footnotes_list[list_index]:
                # empty list -> new paragraph
                list_index += 1
                continue

            if y+self.footnote_height > text_area_height+self.top_margin:
                #print '--', y
                break

            text_width = reduce(lambda a, b: a+b, map(lambda x: x[1], pl), 0)

            if len(pl) <= 1:
                # single word
                space = 0
            elif not self.footnotes_list[list_index+1]:
                # last line in this paragraph
                space = self.footnote_space_width
            else:
                space = (text_area_width-text_width)/(len(pl)-1)

            x = x_offset

            for i in range(len(pl)):

                if i == len(pl)-1 and self.footnotes_list[list_index+1]:
                    # and not index == 0: # and not is_title:
                    # last word in this line
                    x = x_offset+text_area_width-pl[i][1]

                area_window.draw_layout(self.gc, x, y, pl[i][0],
                                        foreground=self._color)

                x += pl[i][1]+space

            if not self.footnotes_list[list_index+1]:
                # last line in this paragraph
                y += self.footnote_par_height
            else:
                y += self.footnote_height
            list_index += 1

        self.footnotes_list = self.footnotes_list[list_index:]



    def create_footnotes_list(self, text_area_width, line_index, begin, end):

        footnotes_list = []

        for f in self.content[line_index].footnotes:
            if begin <= f[1] <= end:
                for par in self.fb.footnotes[f[2]]:

                    index = 0

                    while True:
                        pl, index = self.get_pangolayout_list(text_area_width,
                                                              self.footnote_font,
                                                              par, index)

                        footnotes_list.append(pl)
                        if index == 0:
                            break

                    footnotes_list.append([])

        return footnotes_list


    def get_pangolayout_list(self, text_area_width, font,
                             par, char_index=0, end_char_index=0):

        '''create list:
          [[pangolayout, width, index] [pangolayout, width, index] [...]]
        '''

        def create_pangolayout(word, text_start, text_end):

            pangolayout = self.create_pango_layout(word.encode('utf-8'))

            pangolayout.set_font_description(font)
            pangolayout.set_tabs(self.tab_array)

            attr_list = None

            for start, length, style in par.styles:

                if not start+length <= text_start and not start >= text_end:

                    if not attr_list:
                        attr_list = pango.AttrList()

                    s = start-text_start
                    e = s+length
                    if s < 0: s = 0
                    s = len(word[:s].encode('utf-8'))
                    e = len(word[:e].encode('utf-8'))

                    if style == 'emphasis':
                        attr_list.insert(pango.AttrStyle(pango.STYLE_ITALIC, s, e))
                    elif style == 'strong':
                        attr_list.insert(pango.AttrWeight(pango.WEIGHT_BOLD, s, e))
                    elif style == 'footnote':
                        attr_list.insert(pango.AttrRise(PANGO_SCALE*self.text_height/4, s, e))
                        attr_list.insert(pango.AttrFontDesc(self.footnote_font, s, e))

            if attr_list:
                pangolayout.set_attributes(attr_list)

            return pangolayout


        def split_line(line):
            word = ''
            index = 0
            prev_index = 0
            for c in line:
                index += 1
                if c in string.whitespace:
                    if word:
                        yield (word, prev_index, index)
                        prev_index = index
                        word = ''
                        continue
                #else:
                word += c
            if word: yield (word, prev_index, index+1)
            return


        if end_char_index:
            line = par.data[char_index:end_char_index]
        else:
            line = par.data[char_index:]

        if par.type == 'title':
            hyphanate = 0
        else:
            hyphanate = 1


        text_width = 0
        total_width = 0
        pangolayout_list = []
        new_index = 0

        for word, prev_index, index in split_line(line): #words_list:

            text_start = char_index+prev_index
            text_end = char_index+index
            pangolayout = create_pangolayout(word, text_start, text_end)
            width = pangolayout.get_extents()[1][2]/PANGO_SCALE

            if total_width+width > text_area_width:

                if hyphanate:
                    ## hyphanation

                    lang = self.lang
                    for start, length, style in par.styles:
                        if not start+length <= text_start \
                           and not start >= text_end and style[:4] == 'lang':
                            lang = style[5:]

                    wl = self.hyph.hyphenate(word, lang)

                    for w in wl:
                        text_start = char_index+prev_index
                        text_end = char_index+index
                        pl = create_pangolayout(w+u'-', text_start, text_end)

                        wd = pl.get_extents()[1][2]/PANGO_SCALE
                        if total_width+wd < text_area_width:

                            text_width += wd
                            index -= len(word)-len(w)+1

                            # skip leading '-'
                            if word[len(w)] == '-':
                                index += 1

                            pangolayout_list.append((pl, wd, index))

                            prev_index = index
                            break


                if len(pangolayout_list) == 0:
                    #     
                    for i in range(len(word), 0, -1):
                        w = word[:i]
                        pl = create_pangolayout(w, char_index, char_index+i+1)
                        wd = pl.get_extents()[1][2]/PANGO_SCALE
                        if total_width+wd < text_area_width:
                            text_width += wd
                            prev_index = i
                            pangolayout_list.append((pl, wd, char_index+i))
                            break

                new_index = char_index+prev_index

                break

            text_width += width
            total_width += width+self.space_width
            pangolayout_list.append((pangolayout, width, index))

        return pangolayout_list, new_index


    def calc_back(self):

        def calc_back_1(line_index, index, text_area_width, text_area_height):
            new_line_index = line_index
            #new_index = index
            new_index = 0
            y = 0

            if index:
                # tail of line

                font = self.content[line_index].type in ('title', 'subtitle') \
                       and self.title_font or self.font
                index_list = [0]

                while True:

                    pl, new_index = self.get_pangolayout_list(text_area_width,
                                                              font,
                                                              self.content[line_index],
                                                              new_index, index)

                    if new_index == 0:
                        break
                    index_list.append(new_index)

                index_list.reverse()

                for new_index in index_list:

                    #y += self.text_height

                    if new_index == 0:
                        # new paragraph
                        y += self.text_par_height
                    else:
                        y += self.text_height


                    if y+self.text_height > text_area_height:
                        return new_line_index, new_index

            paragraphs = self.content[:new_line_index]
            paragraphs.reverse()

            for par in paragraphs:
                type = par.type
                #line = par.data

                new_line_index -= 1

                if type == 'image':
                    y += self.text_height
                    #print len(self.fb.images[par.href])

                    pb = self.get_pixbuf(self.fb.images[par.href],
                                         text_area_width)
                    h = pb.get_height()

                    if y+h >= text_area_height:
                        return new_line_index, 0
                    y += h #+self.text_height
                    continue

                font = type in ('title', 'subtitle') and self.title_font or self.font
                index_list = [0]

                new_index = 0
                while True:

                    pl, new_index = self.get_pangolayout_list(text_area_width, font,
                                                              par, new_index)
                    if new_index == 0:
                        break
                    index_list.append(new_index)

                if not index_list:
                    # empty line
                    y += self.text_par_height
                    if y+self.text_height > text_area_height:
                        return new_line_index, 0
                    continue


                index_list.reverse()

                for new_index in index_list:

                    if new_index == 0:
                        # new paragraph
                        y += self.text_par_height
                    else:
                        y += self.text_height

                    if y+self.text_height > text_area_height:
                        return new_line_index, new_index


            return 0, 0

        text_area_height = self._height-self.top_margin-self.bottom_margin

        line_index = self.current_page_line_index
        index = self.current_page_char_index

        if self.single_column:

            #x_offset = self.left_page_left_margin
            #y_offset = self.top_margin
            text_area_width = self._width-self.left_page_left_margin \
                              -self.left_page_right_margin

            line_index, index = calc_back_1(line_index, index,
                                            text_area_width, text_area_height)

        else:
            # right page
            #x_offset = self._width/2+self.right_page_left_margin
            #y_offset = self.top_margin
            text_area_width = self._width/2-self.right_page_left_margin \
                              -self.right_page_right_margin

            line_index, index = calc_back_1(line_index, index,
                                            text_area_width, text_area_height)

            # left page
            #x_offset = self.left_page_left_margin
            #y_offset = self.top_margin
            text_area_width = self._width/2-self.left_page_left_margin \
                              -self.left_page_right_margin

            line_index, index = calc_back_1(line_index, index,
                                            text_area_width, text_area_height)

        return line_index, index


    def go_fore_page(self):
        if len(self.content) <= self.next_page_line_index:
            return

        self.prev_index_list.append((self.current_page_line_index,
                                     self.current_page_char_index))

        #if self.next_page_line_index == len(self.lines):
        #    return
        self.current_page_line_index = self.next_page_line_index
        self.current_page_char_index = self.next_page_char_index
        self.draw_text(go_fore=True)



    def go_back_page(self):
        if self.current_page_line_index == 0 \
           and self.current_page_char_index == 0:
            return
        if self.prev_index_list:
            line_index, index = self.prev_index_list[-1]
            del self.prev_index_list[-1]
        else:
            line_index, index = self.calc_back()

        self.current_page_line_index = line_index
        self.current_page_char_index = index
        self.draw_text()


    def calc_position(self):
        if self.next_page_line_index == len(self.content):
            return 1.0
        if self.text_length:
            cur_pos = reduce(lambda x, y: x+len(y.data),
                             self.content[:self.next_page_line_index], 0)
            return float(cur_pos)/self.text_length

        else:
            return 0.0


    def set_position(self, pos):
        index = self.text_length*pos
        i = 0
        n = 0
        for par in self.content:
            if index <= n:
                break
            n += len(par.data)
            i += 1
        self.current_page_line_index = i
        self.current_page_char_index = 0
        self.draw_text()


    def key_press_event_cb(self, w, event):
        print 'key_press_event_cb'

    def key_release_event_cb(self, w, event):

        if event.keyval in self.keysyms_next:
            self.go_fore_page()
            return True
        if event.keyval in self.keysyms_prev:
            self.go_back_page()
            return True
        return False

    def button_release_event_cb(self, w, event):
        if event.button == 1:
            self.go_fore_page()
        elif event.button == 2:
            #print 'event.button == 2'
            self.go_back_page()


if __name__ == '__main__':
    #import sys
    if len(sys.argv) != 2:
        sys.exit('usage: %s file' % sys.argv[0])

    window = gtk.Window()
    window.connect('destroy', lambda w: gtk.main_quit())
    ob = OrnamentBook(window)
    window.add(ob)

    #text = ''
    #for i in xrange(20):
    #    text = text+str(i)*3+' '
    #text = (text+' aa bb cc ddddd\n')*7
    text = open(sys.argv[1]).read()
    ob.set_text(unicode(text, 'utf-8'))
    ob.realize()
    window.show_all()

    gtk.main()
